/*
 *  Universit degli Studi di Trento - Facolt di scienze
 *
 *  Author:  Paolo Larcheri
 *  Date:    May 2005
 *  
 */
// HCF v3.0
 
 
#include <list>
#include "vbr_traffic.h"
#include "apps/app.h"
#include <math.h>

void dgesl ( double a[], int lda, int n, int ipvt[], double b[], int job );
double dgeco ( double a[], int lda, int n, int ipvt[], double z[] );

using namespace std;

static class VBRTrafficClass : public TclClass {
 public:
   VBRTrafficClass() : TclClass("Application/Traffic/VBR") {}
   TclObject* create(int, const char*const*) {
      return (new VBR_Traffic());
   }
} class_vbr_traffic;


VBR_Traffic::VBR_Traffic() : seqno_(0), n_states(0), curr_state(0), states(0)
{
   bind("packetSize_", &size_);
   bind("maxpkts_", &maxpkts_);
   size_ = 210;
}

VBR_Traffic::~VBR_Traffic()
{
   delete_states();
}

int VBR_Traffic::command(int argc, const char*const* argv)
{
   if (argc == 2)
   {
      if (strcmp(argv[1], "end-config") == 0)
      {
         for (int i = 0; i < n_states; ++i)
            if (states[i]->checkTransitions())
               states[i]->writeable = false; // no further modification after end-config!
            else
            {
               fprintf(stderr, "VBR: please check the assigned transition probability values!\n");
               return TCL_ERROR;
            }
         return TCL_OK;
      }
      else if (strcmp(argv[1], "get-MeanDataRate") == 0)
      {
         Tcl& tcl = Tcl::instance();
         tcl.resultf("%d", computeMeanDataRate());
         return TCL_OK;
      }
      else if (strcmp(argv[1], "get-PeakDataRate") == 0)
      {
         Tcl& tcl = Tcl::instance();
         tcl.resultf("%d", computePeakDataRate());
         return TCL_OK;
      }
   }
   else if (argc == 3)
   {
      if (strcmp(argv[1], "num-states") == 0)
      {
         int n = atoi(argv[2]);
         if (n > 0)
            allocate_states(n);
         return TCL_OK;
      }
   }
   else if (argc == 4)
   {
      if (strcmp(argv[1], "get-EquivDataRate") == 0)
      {
         Tcl& tcl = Tcl::instance();
         double time_sample = atof(argv[2]);
         double prob_bound = atof(argv[3]);
         tcl.resultf("%d", computeEquivDataRate(time_sample, prob_bound));
         return TCL_OK;
      }
   }
   else if (argc == 5)
   {
      if (strcmp(argv[1], "trans-probability") == 0)
      {
         int st1  = atoi(argv[2]);
         int st2  = atoi(argv[3]);
         double prob = atof(argv[4]);
         if ( st1 < 0 || st1 >= n_states || st2 < 0 || st2 >= n_states
              || st1 == st2 || prob <= 0.0 || prob > 1.0 )
         {
               fprintf(stderr, "VBR: please check the assigned transition probability values!\n");
               return TCL_ERROR;
         }
         if (!states[st1]->addTransition(states[st2], prob))
         {
            fprintf(stderr, "VBR: cannot add transition after end-config has been called!\n");
            return TCL_ERROR;
         }
         return TCL_OK;
      }
      else if (strcmp(argv[1], "state-interarrival") == 0)
      {
         int i = atoi(argv[2]);
         double val = atof(argv[4]);
         if (strcmp(argv[3], "constant") == 0)
         {
            states[i]->interval_dist = INTERVAL_CONST;
            states[i]->const_interval = val;
            return TCL_OK;
         }
         else if (strcmp(argv[3], "exponential") == 0)
         {
            states[i]->interval_dist = INTERVAL_EXPO;
            states[i]->mean_interval = val;
            return TCL_OK;
         }
      }
      else if (strcmp(argv[1], "state-packetsize") == 0)
      {
         int i = atoi(argv[2]);
         int val = atoi(argv[4]);
         if (strcmp(argv[3], "exponential") == 0)
         {
            states[i]->pktsize_dist = PKTSIZE_EXPO;
            states[i]->mean_pktsize = val;
            return TCL_OK;
         }
      }
   }
   else if (argc == 6)
   {
      if (strcmp(argv[1], "state-packetsize") == 0)
      {
         int i = atoi(argv[2]);
         int val1 = atoi(argv[4]);
         int val2 = atoi(argv[5]);
         if (strcmp(argv[3], "eq-distributed") == 0)
         {
            if (val1 > VBR_PKTSIZE_THRESHOLD)
               val1 = VBR_PKTSIZE_THRESHOLD;
            if (val2 > VBR_PKTSIZE_THRESHOLD)
               val2 = VBR_PKTSIZE_THRESHOLD;
            states[i]->pktsize_dist = PKTSIZE_EQDIST;
            states[i]->size1 = val1;
            states[i]->size2 = val2;
            return TCL_OK;
         }
      }
   }
   else if (argc == 8)
   {
      if (strcmp(argv[1], "state-packetsize") == 0)
      {
         int i = atoi(argv[2]);
         int val1 = atoi(argv[4]);
         double val2 = atof(argv[5]);
         int val3 = atoi(argv[6]);
         double val4 = atof(argv[7]);
         if (strcmp(argv[3], "two-values") == 0)
         {
            if (val1 > VBR_PKTSIZE_THRESHOLD)
               val1 = VBR_PKTSIZE_THRESHOLD;
            if (val2 > VBR_PKTSIZE_THRESHOLD)
               val2 = VBR_PKTSIZE_THRESHOLD;
            if (VBR_GT_ONE(val2))
            {
               fprintf(stderr, "VBR: probability greater than 1! exiting...");
               return TCL_ERROR;
            }   
            if (!VBR_EQ_ONE(val2+val4))
               val4 = 1 - val2;
            states[i]->pktsize_dist = PKTSIZE_TWOVAL;
            states[i]->size1 = val1;
            states[i]->prob1 = val2;
            states[i]->size2 = val3;
            states[i]->prob2 = val4;
            return TCL_OK;
         }
      }
   }
   return Application::command(argc, argv);
}


void VBR_Traffic::init()
{
   // this avoids to leave size_ and interval_ uniniailized
   next_interval(size_);

   if (agent_)
      agent_->set_pkttype(PT_VBR);
}

void VBR_Traffic::start()
{
    init();
    running_ = 1;
    timeout();
}

double VBR_Traffic::next_interval(int& size)
{
   interval_ = curr_state->getInterval();
   size = size_ = curr_state->getPktSize();
   //fprintf(stderr, "interval: %f, pktsize: %d\n", interval_, size_);
   curr_state = curr_state->getNextState();

   if (++seqno_ < maxpkts_)
      return (interval_);
   else
      return(-1);
}

void VBR_Traffic::allocate_states(int n)
{
   if (states)
      delete_states();

   n_states = n;
   states = new VBR_state*[n];
   for (int i = 0; i < n_states; ++i)
      states[i] = new VBR_state(i, (long)rand());
   curr_state = states[0];
}

void VBR_Traffic::delete_states()
{
   if (!states)
      return;

   for (int i = 0; i < n_states; ++i)
      delete states[i];
   delete [] states;
   states = NULL;
   curr_state = NULL;
   n_states = 0;
}


void VBR_Traffic::computeStationaryProb(double *result)
{
   // Check if there is any absorbent state and notify a WARNING!
   for (int i = 0; i < n_states; ++i)
   {
      if (states[i]->transitions.size() == 1 &&
          states[i]->transitions.front().state->index == i)
        fprintf(stderr, "WARNING: your VBR source has an absorbent state... this can lead to errors while computing MeanDataRate value!\n");
   }

   double linear_system[n_states][n_states+1];

   // a_n.1 = a_n.2 =...= a_n.n = b_n = 1
   for (int j = 0; j <= n_states; ++j)
      linear_system[n_states-1][j] = 1.0;

   // initialize the rest of matrix 
   for (int i = 0; i < n_states-1; ++i)
      for (int j = 0; j <= n_states; ++j)
         linear_system[i][j] = (i == j) ? 1.0 : 0.0;

   // set the values depending on transition probabilities
   for (int i = 0; i < n_states; ++i)
      for (transition_it it = states[i]->transitions.begin(); it != states[i]->transitions.end(); ++it)
      {
         if (it->state->index == n_states-1)
            continue;
         //fprintf(stderr, "%d -> %d   %f\n", i, it->state->index, it->prob);
         linear_system[it->state->index][i] -= it->prob;
      }
   /*for (int i = 0; i < n_states; ++i)
   {
      for (int j = 0; j <= n_states; ++j)
         fprintf(stderr, "%f ", linear_system[i][j]);
      fprintf(stderr, "\n");
   }*/

   // resolve the linear_system by applying GAUSS reduction
   for (int k = 0; k < n_states; ++k)
   {
      if (linear_system[k][k] == 0.0)
      {
         fprintf(stderr, "ERROR: about to execute a division by ZERO... exiting!\n");
         exit(-1);
      }
      for (int i = k+1; i < n_states; ++i)
      {
         // this line is already 0 in column k, so we don't need to touch it
         if (linear_system[i][k] == 0.0)
            continue;

         double coeff = -(linear_system[i][k]/linear_system[k][k]);
         linear_system[i][k] = 0.0; // set to 0 the k-esime column
         for (int j = k+1; j <= n_states; ++j)
         {
            linear_system[i][j] += coeff*linear_system[k][j];
         }
      }
   }


   /*fprintf(stderr, "------------------------------------\n");
   for (int i = 0; i < n_states; ++i)
   {
      for (int j = 0; j <= n_states; ++j)
         fprintf(stderr, "%f ", linear_system[i][j]);
      fprintf(stderr, "\n");
   }*/

   for (int i = n_states-1; i >= 0; --i)
   {
      double tmp = linear_system[i][n_states];
      for (int k = i+1; k < n_states; ++k)
         tmp -= result[k]*linear_system[i][k];
      result[i] = tmp / linear_system[i][i];
   }
 
   /*fprintf(stderr, "------------------------------------\n");
   for (int i = 0; i < n_states; ++i)
      fprintf(stderr, "stationary[%d] = %f\n", i, result[i]);
   fprintf(stderr, "------------------------------------\n");*/
}

int VBR_Traffic::computeMeanDataRate()
{
   if (n_states < 1)
      return -1;
   if (n_states == 1)
      return states[0]->getMeanDataRate();

   double stat_prob[n_states];
   computeStationaryProb(stat_prob);

   int ret = 0;
   double sum = 0.0;
   for (int i = 0; i < n_states; ++i)
   {
      ret += (int)ceil(stat_prob[i] * states[i]->getMeanDataRate());
      //fprintf(stderr, "DataRate        (state %d) = %d\n", i, states[i]->getMeanDataRate());
      //fprintf(stderr, "Stationary Prob (state %d) = %f\n", i, stat_prob[i]);
      //sum += stat_prob[i];
   }
   //fprintf(stderr, "TOT = %f\n", sum);

   return ret;
}


int VBR_Traffic::computePeakDataRate()
{
   int ret = 0;
   for (int i = 0; i < n_states; ++i)
      if (states[i]->getMeanDataRate() > ret)
         ret = states[i]->getMeanDataRate();
   return ret;
}


int VBR_Traffic::computeEquivDataRate(double time_sample, double prob_bound)
{
   list<DFS_entry> feasible_paths;
   double stat_prob[n_states];
   computeStationaryProb(stat_prob);
   double curr_prob = 0.0;
   int mean_data_rate = computeMeanDataRate();

   for (int i = 0; i < n_states; ++i)
      DFS_EquivDataRate(time_sample, 0.0, stat_prob[i], states[i], 0, feasible_paths);

   feasible_paths.sort();

/*   for (list<DFS_entry>::reverse_iterator it = feasible_paths.rbegin(); it != feasible_paths.rend(); ++it)
      fprintf(stderr, "DFS_entry:    %f    %d\n", it->prob, it->bitrate);*/
   for (list<DFS_entry>::reverse_iterator it = feasible_paths.rbegin(); it != feasible_paths.rend(); ++it)
   {
      curr_prob += it->prob;
      if (curr_prob > prob_bound)
         return ((it->bitrate > mean_data_rate) ? it->bitrate : mean_data_rate);
   }

   return mean_data_rate;
}


void VBR_Traffic::DFS_EquivDataRate(double targ_time, double elaps_time, double prob,
                                    VBR_state* state, int bits, list<DFS_entry> &result)
{
   bits += state->getAvgPktSize()*8;

   if (elaps_time >= targ_time)
   {
      elaps_time += state->getAvgInterval();
      result.push_back(DFS_entry(prob, (int)(bits/elaps_time)));
      //fprintf(stderr, "DFS_entry:     %f     %d\n", prob, (int)(bits/elaps_time));
   }
   else
   {
      elaps_time += state->getAvgInterval();
      for (transition_it it = state->transitions.begin(); it != state->transitions.end(); ++it)
         DFS_EquivDataRate(targ_time, elaps_time, prob * it->prob, it->state, bits, result);
   }
}

