# Exclusion of useless packet headers
remove-packet-header AODV aSRM CtrMcast Diffusion Encap\
                HttpInval IMEP IPinIP IVS LDP mcastCtrl MFTP MPLS\
                MIP Ping PGM PGM_SPM PGM_NAK RAP RTP Resv rtProtoDV rtProtoLS SR Src_rt SRM SRMEXT\
                Snoop TCPA TFRC TFRC_ACK TORA GAF UMP Pushback NV

set opt(appORwrt) a                              ;# must append to stats file or trucate it? (w: write, a: append)
set opt(bs)               1  
set opt(nflows)        15
set opt(step)           10
set opt(delta)          0.0015
set opt(seed)           777                          ;# Seed of the Random Number Generator
set opt(x)                560                          ;# X-size of the topography
set opt(y)                560                          ;# Y-size of the topography
set opt(chan)          Channel/WirelessChannel
set opt(netif)          Phy/WirelessPhy
set opt(prop)           Propagation/TwoRayGround
set opt(ant)             Antenna/OmniAntenna
set opt(adhocRouting)   DSDV
set opt(tcp_size)      1500
set opt(mac)            Mac/802_11
set opt(macHC)       Mac/802_11HC
set opt(ifq)               Queue/DropTail/BS
set opt(ifq_len)        50
set opt(num_tc)       4
set opt(ll)                 LL
set opt(nccop)              10
set opt(wtburst)             -1
set opt(bsburst)              -1
set opt(cap_interval)   0.06
set opt(txoplmt)          0.0028  ;# EDCFtime =  0.05 - (0.0028*8) = 0.0276
set opt(stop)            [expr $opt(nflows) * $opt(step)]
set opt(cap_start)    0
set opt(cap_stop)     $opt(stop)
set opt(pcf_start)     0
set opt(pcf_stop)      $opt(stop)
set num_n                $opt(nflows)

source "init.tcl"
source "stats.tcl"

set datarate_fast   11e6      ;# fast data transmission rate (11Mbps)
set datarate_slow  2e6     ;# slow data transmission rate ( 2Mbps)

Queue/DropTail set drop_front_               false           ;# Drop-from-front (rather than from tail)
Queue/DropTail set queue_in_bytes_      false           ;# Queue measured in bytes?
Queue/DropTail/BS set queue_in_byte_  false           ;# Queue measured in bytes?
Phy/WirelessPhy set bandwidth_  11e6    ;# Bit rate
Mac/802_11   set basicRate_     1e6
Mac/802_11   set dataRate_      11e6
Mac/802_11HC set basicRate_     1e6
Mac/802_11HC set dataRate_      11e6
Agent/DSDV set perup_ [expr $opt(stop) * 2]   ;# update period
Agent/TCP set packetSize_ $opt(tcp_size)


# Creating simulator instance and setting up other stuff
set ns_ [new Simulator]
set topo        [new Topography]
$ns_ use-scheduler Heap   ;# The scheduler uses a heap structure
ns-random seed  $opt(seed)
open_stat_files
$ns_ use-newtrace
set tracefd     [open prova.tr w]
$ns_ trace-all $tracefd

# Setting up hierarchical address
$ns_ set-address-format hierarchical           ;# Wired-cum-wireless scenarios need hierarchical routing
$ns_ node-config -addressType hierarchical
AddrParams set domain_num_               2                               ;# Number of domains
lappend cluster_num                             2 1                             ;# Number of clusters in each domain
AddrParams set cluster_num_ $cluster_num
lappend eilastlevel                             1 1 [expr $opt(nflows) + 1]    ;# Number of nodes in each cluster
AddrParams set nodes_num_ $eilastlevel

# GOD & TOPOGRAPHY
$topo load_flatgrid $opt(x) $opt(y)
create-god [expr $num_n + 1]

# ------------------
# Wired nodes
# ------------------

set remote_node  [$ns_ node 0.0.0]
set gateway_node [$ns_ node 0.1.0]

#                   OPTION TYPE     DESCRIPTION & VALUE USED IN THE SIMULATION
#                  -------------   ----------------------------------------------
#
# $ns_ node-config -adhocRouting    Routing Agent: DSDV
#                  -llType          LLC sublayer: LL
#                  -ifqType         Interface Queue: Queue/DropTail/BS
#                  -ifqLen          Max. Interface Queue size: configurable
#                  -macType         MAC sublayer: Mac/802_11HC (for Hybrid Coordinator)
#                                   MAC sublayer: Mac/802_11   (for Wireless STAs)
#                  -antType         Antenna type: Antenna/OmniAntenna
#                  -propInstance    Radio Propagation model: Propagation/TwoRayGround
#                  -phyType         Network Interface: Phy/WirelessPhy
#                  -channel         Channel type: Channel/WirelessChannel
#                  -topoInstance    Topography: $topo
#                  -wiredRouting    Wired Routing: ON for BS, OFF for WSs
#                  -agentTrace      Agent tracing: OFF
#                  -routerTrace     Routing tracing: OFF
#                  -macTrace        MAC tracing: OFF
#                  -IncomingErrProc Error Model: ChnErrProc

#----------------------------------------------------------------------------------------

$ns_ node-config \
                        -adhocRouting      $opt(adhocRouting) \
                        -llType                   $opt(ll) \
                        -ifqType                 $opt(ifq) \
                        -ifqLen                   $opt(ifq_len) \
                        -macType              $opt(macHC) \
                        -antType                $opt(ant) \
                        -propInstance         [new $opt(prop)] \
                        -phyType                $opt(netif) \
                        -channel                 [new $opt(chan)] \
                        -topoInstance           $topo \
                        -wiredRouting          ON \
                        -agentTrace              ON \
                        -routerTrace             OFF \
                        -macTrace                 ON 

# Configures TSPECS for each category
source "tc_par.tcl"


#----------------------------------------------------------------------------------------
#                    AccessPoint Configuration
#----------------------------------------------------------------------------------------
set BS [$ns_ node 1.0.0]
classify $BS 0 "BS"

# Activation/Deactivation of bursting transmission capability
$obj(ifq,0) init-llc -1         ;# possible values: 0, -1, #n#

# Informs the BS about the size of voice packets (bytes) and
# framing time (secs) according to the selected vocoder (see param.tcl)
$obj(mac,0) set-vocoder 320  0.01

# Sets the number of Access Categories and the TXOPlimit value (EDCA)
$obj(mac,0) init-mac $opt(num_tc) $opt(txoplmt)

# For each AccessCategory...
for {set tc 0} {$tc < $opt(num_tc)} {incr tc} {
   # Adds a new queue
   $obj(ifq,0) add-queue $tc $tc_par($tc,tout) $opt(ifq_len)
   # Sets the QoS parameters of the MAC sublayer
   $obj(mac,0) set-qos $tc $tc_par($tc,aifs) $tc_par($tc,cwmin) $tc_par($tc,srl) $tc_par($tc,lrl) $tc_par($tc,cwmax)
}

if {$opt(cap_start) >= 0} {
                # Sets the Controlled Access Period interval
                # and the number of CCOPs during each CCI
                $obj(mac,0) set-cap $opt(cap_interval) $opt(nccop)
                $ns_ at $opt(cap_start) "$obj(mac,0) cap-start"         ;# Starts the CAP mechanism at $opt(cap_start)
                $ns_ at $opt(cap_stop) "$obj(mac,0) cap-stop"           ;# Stops the CAP mechanism at $opt(cap_stop)
}
if {$opt(pcf_start) >= 0} {
           $ns_ at $opt(pcf_start) "$obj(mac,0) pcf-start"         ;# Starts the PCF mechanism at $opt(pcf_start)
           $ns_ at $opt(pcf_stop) "$obj(mac,0) pcf-stop"           ;# Stops the PCF mechanism at $opt(pcf_stop)
}


#create links between wired nodes and the base station
  $ns_ duplex-link $remote_node $gateway_node 1000Mb 10ms DropTail
  $ns_ queue-limit $remote_node $gateway_node 100000
  $ns_ duplex-link $gateway_node $BS 100Mb 5us DropTail
  $ns_ duplex-link-op $remote_node $gateway_node orient down
  $ns_ duplex-link-op $gateway_node $BS orient left-down



#----------------------------------------------------------------------------------------
#                    Wireless Nodes Configuration
#----------------------------------------------------------------------------------------
$ns_ node-config        -macType        $opt(mac) \
                                    -wiredRouting   OFF             ;# Deactivation of wired routing

for {set j 1} {$j <= $num_n} {incr j} {
    set node($j) [ $ns_ node 1.0.$j ]
    $node($j) base-station [AddrParams addr2id [$BS node-addr]]
    $node($j) random-motion 0
    $node($j) set X_ 1.0
    $node($j) set Y_ 0.0
    $node($j) set Z_ 0.0
    $ns_ initial_node_pos $node($j) 15

    classify $node($j) $j "WN$j"
    # Communicates the datarate to the station
    $obj(mac,$j) set-datarate $datarate_fast
    # Communicates the datarate to the BS
    $obj(mac,0) STA_datarate [$obj(mac,$j) mac-addr] $datarate_fast

    # Activation/Deactivation of bursting transmission capability
    $obj(ifq,$j) init-llc -1                ;# possible values: 0, -1, #n#
    # Sets the number of Access Categories and the TXOPlimit value
    $obj(mac,$j) init-mac $opt(num_tc) $opt(txoplmt)   

     for {set tc 0} {$tc < $opt(num_tc)} {incr tc} {
           # Adds a new queue for each Access Category
           $obj(ifq,$j) add-queue $tc $tc_par($tc,tout) $opt(ifq_len)
           # Sets the QoS parameters of the MAC sublayer
           $obj(mac,$j) set-qos $tc $tc_par($tc,aifs) $tc_par($tc,cwmin) \
                                               $tc_par($tc,srl) $tc_par($tc,lrl) $tc_par($tc,cwmax)
     }

     # This is a UDP uplink station:
     # inserts Multimedia Access Category 2 of the node
     #$ns_ at [expr ($j-1)*$opt(step)] "$obj(mac,0) cf-poll [$obj(mac,$j) mac-addr] 2"
     $obj(mac,0) cf-poll [$obj(mac,$j) mac-addr] 2
}


# Informs the HC about fast and slow data rates (see param.tcl),
# about the number of stations with slow datarate and about the
# TCP data packet size (in bytes) so that it can calculate the X
# parameter, the Beacon interval duration and each station's CFP
# TXOP limit for the current network configuration
$obj(mac,0) evalX $datarate_fast $datarate_slow 0 $opt(tcp_size)


#----------------------------------------------------------------------------------------
#                    Traffic Configuration
#----------------------------------------------------------------------------------------
# setting up streams
  for {set j 1} {$j <= $opt(nflows)} {incr j} {
     global udp_down udp_up null_up null_down
      # DOWN stream
     set udp_down($j) [new Agent/UDP]
     $udp_down($j) set fid_ $j                   ;# flow ID
     set null_down($j) [new Agent/Null]
     $null_down($j) set fid_ $j                   ;# flow ID
     $null_down($j) set jitter_id 1 ;# used for differentiating avg_delay/jitter statististics (1 downlink, 2 uplink)
     $ns_ attach-agent $remote_node $udp_down($j)
     $ns_ attach-agent $node($j) $null_down($j)
     $ns_ connect $udp_down($j) $null_down($j)
     set cbr_down($j) [new Application/Traffic/CBR]
     $cbr_down($j) attach-agent $udp_down($j)
     $cbr_down($j) set packetSize_ 160   ;# equal to
     $cbr_down($j) set interval_ 0.01       ;# 128 Kb/s.
     # UP stream
     set udp_up($j) [new Agent/UDP]
     $udp_up($j) set fid_ [expr $j+$opt(nflows)]                   ;# flow ID
     set null_up($j) [new Agent/Null]
     $null_up($j) set fid_ [expr $j+$opt(nflows)]                    ;# flow ID
     $null_up($j) set jitter_id 2 ;# used for differentiating avg_delay/jitter statististics (1 downlink, 2 uplink)
     $ns_ attach-agent $node($j) $udp_up($j)
     $ns_ attach-agent $remote_node $null_up($j)
     $ns_ connect $udp_up($j) $null_up($j)
     set cbr_up($j) [new Application/Traffic/CBR]
     $cbr_up($j) attach-agent $udp_up($j)
     $cbr_up($j) set packetSize_ 160   ;# equal to
     $cbr_up($j) set interval_ 0.01       ;# 128 Kb/s.

     set time [expr $opt(step) * ($j-1)+1]
     $ns_ at $time " puts \"time  $time -> starting bidirectional stream $j \""
     $ns_ at [expr $time + $j * $opt(delta)] "$cbr_down($j) start"
     $ns_ at [expr $time + $j * $opt(delta)] "$cbr_up($j) start"
     $ns_ at [expr $time+$opt(step)] "stats $j"
     $ns_ at [expr $opt(stop)+1] "$cbr_down($j) stop"
     $ns_ at [expr $opt(stop)+1] "$cbr_up($j) stop"
}

  proc on_halt {} {
      global ns_ tracefd opt num_n
      puts "time $opt(stop) -> NS EXITING..."
      close $tracefd
      close_stat_files
      $ns_ halt
  }

  puts "----------------------------------"
  puts "Starting Simulation..."
  puts "----------------------------------"
  puts "Addresses:"
  puts "remote node - [$remote_node node-addr]"
  puts "gateway     - [$gateway_node node-addr]"
  puts "BS          - [$BS node-addr]"
  puts "----------------------------------"


  $ns_ at [expr $opt(stop)+2] "on_halt"
  $ns_ run
