###################################################################
#                                                                                                                                                                                       #
# finish: Called at the end of the simulation, it produces the                                                                                    #
#         output file by calling others subroutines                                                                                                          #
#                                                                                                                                                                                       #
###################################################################


proc open_stat_files {} {
    global dd_filed ud_filed ab_filed
    set dd_filed [open "drop_down.txt" w]
    set ud_filed [open "drop_up.txt" w]
    set ab_filed [open "avg_buff.txt" w]
}

proc close_stat_files {} {
    global dd_filed ud_filed ab_filed
    close $dd_filed
    close $ud_filed
    close $ab_filed
}


proc stats {num} {
    global dd_filed ud_filed ab_filed
    dDROP_stat $num $dd_filed
    uDROP_stat $num $ud_filed
    AVGBUFF_stat $num $ab_filed
}

# Init global statistics variable
set prev_dpkt_sent 0
set prev_dpkt_drop 0
set prev_upkt_sent 0
set prev_upkt_drop 0

#
# Print to file the ratio of the number of dropped packets to the number of sent packets (downlink)
#
proc dDROP_stat {num filed} {
    global udp_down prev_dpkt_sent prev_dpkt_drop obj

    set pkt_sent 0
    for {set i 1} {$i <= $num} {incr i} {
         set pkt_sent [expr $pkt_sent + [$udp_down($i) set seqno_]]
    }

    set pkt_sent [expr $pkt_sent - $prev_dpkt_sent ]
    set prev_dpkt_sent [expr $prev_dpkt_sent + $pkt_sent]
    
    set tot_pkt_drop [BS_udp_pkt_drop]
    set pkt_drop [expr $tot_pkt_drop - $prev_dpkt_drop]
    set prev_dpkt_drop $tot_pkt_drop

    #puts $filed "$num   $pkt_sent  $pkt_drop   [expr $pkt_drop.0/$pkt_sent.0]"
    puts $filed "$num    [expr $pkt_drop.0/$pkt_sent.0]"
}

#
# Print to file the ratio of the number of dropped packets to the number of sent packets (uplink)
#
proc uDROP_stat {num filed} {
    global udp_up prev_upkt_sent prev_upkt_drop obj

    set pkt_sent 0
    for {set i 1} {$i <= $num} {incr i} {
         set pkt_sent [expr $pkt_sent + [$udp_up($i) set seqno_]]
    }

    set pkt_sent [expr $pkt_sent - $prev_upkt_sent]
    set prev_upkt_sent [expr $prev_upkt_sent + $pkt_sent]

    set tot_pkt_drop [WN_udp_pkt_drop $num]
    set pkt_drop [expr $tot_pkt_drop - $prev_upkt_drop]
    set prev_upkt_drop $tot_pkt_drop
    
    #puts $filed "$num   $pkt_sent  $pkt_drop   [expr $pkt_drop.0/$pkt_sent.0]"
    puts $filed "$num     [expr $pkt_drop.0/$pkt_sent.0]"
}

#
# print the avg buffer size of the BS
#
proc AVGBUFF_stat {num filed} {
      global obj curr_queue_stat prev_avg_buff
      $obj(ifq,0) cmd dump_avg_size             ;# calling C++ method for getting the avg_size
      puts $filed "$num  $curr_queue_stat(ifq,avg_size,3)"
      $obj(ifq,0) cmd reset_avg_size             ;# calling C++ method for resetting avg_size counting
}

proc BS_udp_pkt_drop {} {
        global curr_queue_stat obj
        $obj(ifq,0) cmd dump_queue               ;# getting info from BS queues (calling C++ method)
        set total_drop [expr $curr_queue_stat(ifq,drop,3,udp) + $curr_queue_stat(ifq,tout,3,udp)]
        return $total_drop
}

proc WN_udp_pkt_drop {num} {
        global curr_queue_stat opt num_n obj opt
        set total_drop 0
        for {set i 1} {$i <= $num} {incr i} {
             $obj(ifq,$i) cmd dump_queue           ;# getting info from wireless-node queues (calling C++ method)
             set total_drop [expr $total_drop + $curr_queue_stat(ifq,drop,2,udp) + $curr_queue_stat(ifq,tout,2,udp)]
        }
        return $total_drop 
}



######################################################################
#                                                                    #
# queue_stat: called by drop-tail.cc to dump Interface Queue           #
#           statistics                                               #
#                                                                    #
#                n: queue identifier                                 #
#           q_arrv: total number of packet arrival                   #
#           q_size: average queue size                               #
#            rcv_*: packets received from the LLC                    #
#           drop_*: packets discarded due to saturation              #
#                                                                    #
#           * rapresents the packet type                             #
#                                                                    #
######################################################################

proc queue_stat {n limit q_arrv q_size rcv_tcp rcv_ack rcv_message rcv_ARP rcv_udp rcv_voip \
                drop_tcp drop_ack drop_message drop_ARP drop_udp drop_voip \
                tout_tcp tout_ack tout_message tout_ARP tout_udp tout_voip} {
        global curr_queue_stat
        set curr_queue_stat(bs_q_size,$n) $q_size
        set curr_queue_stat(bs_q_arrv,$n) $q_arrv
        set curr_queue_stat(bs_q_limit,$n) $limit
        foreach tip {rcv drop tout} {
                foreach typ {ARP message ack tcp udp voip} {
                        set curr_queue_stat(ifq,$tip,$n,$typ) [set $tip\_$typ]
                }
        }
}

######################################################
#                                                                                                                                                  #
#  queue_avgsize: called by c++ class BS_DropTail (see queue/drop-tail.cc) to dump     #
#                             the average queue size.                                                                            #
#  note: the avg counting is resetted                                                                                      #
#                                                                                                                                                  #
######################################################

proc queue_avgsize {n avg_size} {
      global curr_queue_stat
      puts "$n -> $avg_size"
      set curr_queue_stat(ifq,avg_size,$n) $avg_size
}